//
// This program and source are subject to the GPL Open Source License
// Please refer to the following document
// http://www.borland.com/kylix/gpl.html

//---------------------------------------------------------------------------
#include <clx.h>
#pragma hdrstop

#include "SMP_Test_Form.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef __linux__
#include <sys/resource.h>
#include <sched.h>
#include <sys/mman.h>
#endif
#include "SMP_ATask.h"
#include "SMP_CTask.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.xfm"
TSMPTestForm *SMPTestForm;
//---------------------------------------------------------------------------
__fastcall TSMPTestForm::TSMPTestForm(TComponent* Owner)
  : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::FormShow(TObject *Sender)
{
  #ifdef __linux__
  TSchedParam sp;

  def_prio=sched_get_priority_min(SCHED_OTHER);
  max_prio=sched_get_priority_max(SCHED_RR);
  min_prio=sched_get_priority_min(SCHED_RR);

  sp.sched_priority=def_prio;
  sched_setscheduler(0,SCHED_OTHER,&sp);
  setpriority(PRIO_PROCESS,getpid(),-20);
  #endif

  #ifdef __WIN32__
  SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_NORMAL);

  // N/A in Windows
  LockMemory->Checked=false;
  LockMemory->Enabled=false;
  #endif

  // disable realtime to avoid lockups
  RunRealtime->Checked=false;
  RunRealtime->Enabled=false;
  AdjustTasks(RunTasks->ItemIndex+1);
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::FormClose(TObject *Sender, TCloseAction &Action)
{
	AdjustTasks(0);
  Timer1->Enabled=false;
}
//---------------------------------------------------------------------------
void TSMPTestForm::SetControls(bool enabled)
{
  Timer1->Enabled=enabled;
  RunTasks->Enabled=enabled;
  CloseButton->Enabled=enabled;
  CodeType->Enabled=enabled;
  Operation->Enabled=enabled;
  MemorySize->Enabled=enabled;
  RunRealtime->Enabled=enabled;
  LockMemory->Enabled=enabled;

  if (enabled)
  {
    // enable at your own risk
    RunRealtime->Enabled=false;

    #ifndef __linux__
    LockMemory->Enabled=false;
    #endif
  }
}
//---------------------------------------------------------------------------
void TSMPTestForm::AdjustTasks(int NTasks)
{
	if ((NTasks!=RTasks) && RunTasks->Enabled)
  {
  	SetControls(false);

    // cancel all running threads

		for (int i=0;i<NSMPTASKS;i++)
		if (Tasks[i]!=NULL)
    {
			Tasks[i]->Terminate();
			Tasks[i]->WaitFor();
			delete Tasks[i];
			Tasks[i]=NULL;
		}

		RTasks=NTasks;

		if (RTasks>0)
    {
    	// start only one thread for calibration

			if (CodeType->ItemIndex==0)
			Tasks[0]=new TSMP_AThread();
			else
			Tasks[0]=new TSMP_CThread();

      if (RunRealtime->Checked)
      {
        #ifdef __linux__
        Tasks[0]->Resume();
        TSchedParam sp;
        sp.sched_priority=min_prio;
        pthread_setschedparam(Tasks[0]->ThreadID,SCHED_RR,&sp);
        #endif

        #ifdef __WIN32__
        Tasks[0]->Priority=tpNormal;
        Tasks[0]->Resume();
        #endif
      }
      else
      {
        #ifdef __linux__
        Tasks[0]->Resume();
        TSchedParam sp;
        sp.sched_priority=def_prio;
        pthread_setschedparam(Tasks[0]->ThreadID,SCHED_OTHER,&sp);
        #endif

        #ifdef __WIN32__
        Tasks[0]->Priority=tpLower;
        Tasks[0]->Resume();
        #endif
      }

			Timer1->Enabled=true;

      double dt,pc;
			double LastTR=0.0;
			int c=0;

      // this loop calibrates a single thread to 100%
      // it is best to not move the mouse during this process

			do
      {
				TotalRun=0.0;
				while (TotalRun==0.0)
        {
          // sleep this thread so we don't skew results
          Sleep(50);
					Application->ProcessMessages();
				}
				dt=fabs(TotalRun-LastTR);
				LastTR=TotalRun;
				if ((c>=3) && (TotalRun>0.0))
				pc=dt/TotalRun;
				else
				pc=1.0;
				c++;
			}
      while ((c<=10) && (pc>0.01));

			TotalOne=TotalRun;
			ProgressBar1->Max=TotalOne*NSMPTASKS;

      // now start up the rest of the threads

			for (int i=1;i<RTasks;i++)
      {
				if (CodeType->ItemIndex==0)
				Tasks[i]=new TSMP_AThread();
				else
				Tasks[i]=new TSMP_CThread();

        if (RunRealtime->Checked)
        {
          #ifdef __linux__
          Tasks[i]->Resume();
          TSchedParam sp;
          sp.sched_priority=min_prio;
          pthread_setschedparam(Tasks[i]->ThreadID,SCHED_RR,&sp);
          #endif

          #ifdef __WIN32__
          Tasks[i]->Priority=tpNormal;
          Tasks[i]->Resume();
          #endif
        }
        else
        {
          #ifdef __linux__
          Tasks[i]->Resume();
          TSchedParam sp;
          sp.sched_priority=def_prio;
          pthread_setschedparam(Tasks[i]->ThreadID,SCHED_OTHER,&sp);
          #endif

          #ifdef __WIN32__
          Tasks[i]->Priority=tpLower;
          Tasks[i]->Resume();
          #endif
        }
			}
		}

    SetControls(true);

		AnsiString s="Run "+IntToStr(RTasks)+" Thread";
    if (RTasks>1) s=s+"s";
		RunTasks->Caption=s;
	}
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::Timer1Timer(TObject *Sender)
{
  // avoid multiple passes (should never happen - but just in case)
  static bool InTimer=false;
  if (InTimer) return;
  InTimer=true;

	double PCent=0.0;
	double t=0.0;
	double bps=0.0;

	for (int i=0;i<RTasks;i++)
	if (Tasks[i]!=NULL)
  {
		if (CodeType->ItemIndex==0)
    {
			t+=((TSMP_AThread*)Tasks[i])->NumRun;
			bps+=((TSMP_AThread*)Tasks[i])->BytesPerSecond;
		} else {
			t+=((TSMP_CThread*)Tasks[i])->NumRun;
			bps+=((TSMP_CThread*)Tasks[i])->BytesPerSecond;
		}
	}
	TotalRun=t;
	TotalBytesPerSecond=bps;

	ProgressBar1->Position=t;
	if (TotalOne>0.0) PCent=100.0*TotalRun/TotalOne;
	if (RunTasks->Enabled)
  {
		Edit1->Text=FloatToStrF(t,ffNumber,12,0)
      +" = "+FloatToStrF(PCent,ffFixed,3,0)+"%";
		Edit2->Text=FloatToStrF(TotalBytesPerSecond,ffNumber,12,0)
			+" = "+FloatToStrF(TotalBytesPerSecond/1024/1024,ffNumber,12,1)+" MB/S";
	} else {
		Edit1->Text="Please wait...";
		Edit2->Text="Please wait...";
	}

  InTimer=false;
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::RunTasksClick(TObject *Sender)
{
  AdjustTasks(0);
  AdjustTasks(RunTasks->ItemIndex+1);
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::RunRealtimeClick(TObject *Sender)
{
  #ifdef __linux__
  TSchedParam sp;
  if (RunRealtime->Checked)
  {
    sp.sched_priority=min_prio+1; // one higher to retain control
    sched_setscheduler(0,SCHED_RR,&sp);
    for (int i=0;i<RTasks;i++)
    {
      sp.sched_priority=min_prio;
      pthread_setschedparam(Tasks[i]->ThreadID,SCHED_RR,&sp);
    }
  }
  else
  {
    // lower tasks first or we lose control
    for (int i=0;i<RTasks;i++)
    {
      sp.sched_priority=def_prio;
      pthread_setschedparam(Tasks[i]->ThreadID,SCHED_OTHER,&sp);
    }
    sp.sched_priority=def_prio;
    sched_setscheduler(0,SCHED_OTHER,&sp);
    setpriority(PRIO_PROCESS,getpid(),-20);
  }
  #endif

  #ifdef __WIN32__
  if (RunRealtime->Checked)
  {
    SetPriorityClass(GetCurrentProcess(),REALTIME_PRIORITY_CLASS);
    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_TIME_CRITICAL);
    for (int i=0;i<RTasks;i++)
    {
      Tasks[i]->Priority=tpNormal;
    }
  }
  else
  {
    for (int i=0;i<RTasks;i++)
    {
      Tasks[i]->Priority=tpLower;
    }
    SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);
    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_NORMAL);
  }
  #endif
}
//---------------------------------------------------------------------------
void __fastcall TSMPTestForm::LockMemoryClick(TObject *Sender)
{
  #ifdef __linux__
  if (LockMemory->Checked) mlockall(MCL_CURRENT | MCL_FUTURE);
  else munlockall();
  #endif

  #ifdef __WIN32__
  // not possible
  #endif
}
//---------------------------------------------------------------------------

